from extendedstim.Code.QuantumCode.MajoranaCode import MajoranaCode
from extendedstim.Circuit.Circuit import Circuit
from extendedstim.Code.QuantumCode.MajoranaCSSCode import MajoranaCSSCode
from extendedstim.Code.QuantumCode.PauliCSSCode import PauliCSSCode
from extendedstim.Code.QuantumCode.PauliCode import PauliCode
from extendedstim.tools.TypingTools import isinteger, isreal


#%%  USER：===将量子码转换为量子线路===
def Code2Circuit(code,p_noise,p_measure,noise_model,cycle_number):
    ##  ---数据预处理---
    assert isreal(p_noise) and 0<=p_noise<=1
    assert isreal(p_measure) and 0<=p_measure<=1
    assert isinteger(cycle_number) and cycle_number>=0
    assert isinstance(code,MajoranaCode) or isinstance(code,PauliCode)

    ##  ---根据量子码类型选择不同的处理函数---
    ##  处理现象级噪声
    if noise_model=='phenomenological':
        if isinstance(code,MajoranaCSSCode):
            return MajoranaCSSCode2PhenomenologicalCircuit(code,p_noise,p_measure,cycle_number)
        elif isinstance(code,PauliCSSCode):
            return PauliCSSCode2PhenomenologicalCircuit(code,p_noise,p_measure,cycle_number)
        elif isinstance(code,MajoranaCode):
            raise NotImplementedError
        elif isinstance(code,PauliCode):
            raise NotImplementedError
        else:
            raise NotImplementedError

    ##  处理电路级噪声
    elif noise_model=='circuit-level':
        if isinstance(code,MajoranaCSSCode):
            return MajoranaCSSCode2CircuitLevelCircuit(code,p_noise,p_measure,cycle_number)
        elif isinstance(code,PauliCSSCode):
            return PauliCSSCode2CircuitLevelCircuit(code,p_noise,p_measure,cycle_number)
        elif isinstance(code,MajoranaCode):
            raise NotImplementedError
        elif isinstance(code,PauliCode):
            raise NotImplementedError
        else:
            raise NotImplementedError

    ##  处理代码容量级噪声
    elif noise_model=='code-capacity':
        if isinstance(code,MajoranaCSSCode):
            return MajoranaCSSCode2CodeCapacityCircuit(code,p_noise,p_measure,cycle_number)
        elif isinstance(code,PauliCSSCode):
            return PauliCSSCode2CodeCapacityCircuit(code,p_noise,p_measure,cycle_number)
        elif isinstance(code,MajoranaCode):
            raise NotImplementedError
        elif isinstance(code,PauliCode):
            raise NotImplementedError
        else:
            raise NotImplementedError

    ##  其他类型抛出异常
    else:
        raise ValueError('noise_model must be phenomenological, circuit-level, or code-capacity')


#%%  KEY：===将Majorana CSS code转换为现象级噪声下的测试线路===
def MajoranaCSSCode2PhenomenologicalCircuit(code,p_noise,p_measure,cycle_number:int):

    ##  ---数据预处理---
    assert isinstance(code,MajoranaCSSCode)
    stabilizers_x=code.generators_x
    stabilizers_z=code.generators_z
    logical_x=code.logical_operators_x
    logical_z=code.logical_operators_z

    ##  获取logical operators
    logical_occupy=[]
    for i in range(len(logical_x)):
        logical_occupy.append(1j*logical_x[i]@logical_z[i])

    majorana_number=code.physical_number
    stabilizer_number = len(stabilizers_x) + len(stabilizers_z)

    ##  ---生成线路---
    ##  初始化
    circuit = Circuit()
    circuit.append({'name':'FR','target':range(code.physical_number)})

    ##  第一轮测量假设完美的初始化
    observable_include= []  # 记录可观测量的索引
    for i,logical_operator in enumerate(logical_occupy):
        circuit.append({"name":"MPP","target":logical_operator})
        observable_include.append(len(circuit._measurements)-1)
    for i,stabilizer in enumerate(stabilizers_x):
        circuit.append({"name":"MPP","target":stabilizer})
    for i,stabilizer in enumerate(stabilizers_z):
        circuit.append({"name":"MPP","target":stabilizer})

    ##  循环多轮，多轮测量错误与量子噪声信道
    for _ in range(cycle_number):

        ##  量子噪声信道
        for i in range(majorana_number):
            circuit.append({"name":"FDEPOLARIZE1","target":i,"p":p_noise})

        ##  测量稳定子
        for i,stabilizer in enumerate(stabilizers_x):
            circuit.append({"name":"MPP","target":stabilizer,'p':p_measure})
        for i,stabilizer in enumerate(stabilizers_z):
            circuit.append({"name":"MPP","target":stabilizer,'p':p_measure})

        ##  添加探测器
        for i in range(stabilizer_number):
            circuit.append({"name":"DETECTOR","target":[-i - 1, -i - stabilizer_number-1]})

    ##  最后一轮测量稳定子，假设没有噪声
    for i,stabilizer in enumerate(stabilizers_x):
        circuit.append({"name":"MPP","target":stabilizer})
    for i,stabilizer in enumerate(stabilizers_z):
        circuit.append({"name":"MPP","target":stabilizer})

    ##  测量逻辑算符
    for i,logical_operator in enumerate(logical_occupy):
        circuit.append({"name":"MPP","target":logical_operator})
        circuit.append({"name":"OBSERVABLE_INCLUDE","target":[len(circuit._measurements)-1, observable_include[i]]})

    ##  ---返回线路---
    return circuit


#%%  KEY：===将Majorana CSS code转换为线路级噪声下的测试线路===
def MajoranaCSSCode2CircuitLevelCircuit(code,p_noise,p_measure,cycle_number:int):

    ##  ---数据预处理---
    ##  获取稳定子
    stabilizers_x=code.generators_x
    stabilizers_z=code.generators_z

    ##  获取逻辑算符
    logical_x=code.logical_operators_x
    logical_z=code.logical_operators_z
    logical_occupy=[1j*logical_x[temp]@logical_z[temp] for temp in range(len(logical_x))]  # 粒子数算符组作为逻辑算符组

    ##  获取数目
    majorana_number=code.physical_number
    stabilizer_number = len(stabilizers_x) + len(stabilizers_z)
    pauli_number=stabilizer_number

    ##  ---生成线路---
    ##  初始化
    circuit = Circuit()
    circuit.append({'name':'FR','target':range(majorana_number)})
    circuit.append({'name':'R','target':range(pauli_number)})

    ##  第一轮测量假设完美的初始化
    observable_include= []  # 记录可观测量的索引
    for i,logical_operator in enumerate(logical_occupy):
        circuit.append({"name":"MPP","target":logical_operator})
        observable_include.append(len(circuit._measurements)-1)
    for i,stabilizer in enumerate(stabilizers_x):
        circuit.append({"name":"MPP","target":stabilizer})
    for i,stabilizer in enumerate(stabilizers_z):
        circuit.append({"name":"MPP","target":stabilizer})

    ##  添加第一轮噪声
    for i in range(majorana_number):
        circuit.append({"name": "FDEPOLARIZE1", "target": i, "p": p_noise})
    for i in range(pauli_number):
        circuit.append({"name": "DEPOLARIZE1", "target": i, "p": p_noise})

    ##  循环多轮，多轮测量错误与量子噪声信道
    for _ in range(cycle_number):

        ##  添加稳定子测量
        for i,stabilizer in enumerate(stabilizers_x):
            sequence_temp=syndrome_majorana_css_measurement_circuit(stabilizer, i, 'x', p_noise, p_measure)
            for temp in sequence_temp:
                circuit.append(temp)
        for i,stabilizer in enumerate(stabilizers_z):
            sequence_temp=syndrome_majorana_css_measurement_circuit(stabilizer, i + len(stabilizers_x), 'z', p_noise, p_measure)
            for temp in sequence_temp:
                circuit.append(temp)

        ##  添加检测器
        for i in range(stabilizer_number):
            circuit.append({"name":"DETECTOR","target":[-i - 1, -i - stabilizer_number-1]})

    ##  最后一轮测量假设是没有噪声的
    ##  添加稳定子测量
    for i,stabilizer in enumerate(stabilizers_x):
        circuit.append({"name":"MPP","target":stabilizer})
    for i,stabilizer in enumerate(stabilizers_z):
        circuit.append({"name":"MPP","target":stabilizer})

    ##  添加检测器
    for i in range(stabilizer_number):
        circuit.append({"name": "DETECTOR", "target": [-i - 1, -i - stabilizer_number - 1]})

    ##  测量逻辑算符
    for i,logical_operator in enumerate(logical_occupy):
        circuit.append({"name":"MPP","target":logical_operator})
        circuit.append({"name":"OBSERVABLE_INCLUDE","target":[len(circuit._measurements)-1, observable_include[i]]})

    ##  ---返回线路---
    return circuit


#%%  KEY：将Pauli CSS code转换为现象级噪声下的测试线路
def PauliCSSCode2PhenomenologicalCircuit(code,p_noise,p_measure,cycle_number:int):
    pass


#%%  KEY：===将Pauli CSS code转换为电路级噪声下的测试线路===
def PauliCSSCode2CircuitLevelCircuit(code,p_noise,p_measure,cycle_number:int):
    ##  ---数据预处理---
    ##  获取稳定子
    stabilizers_x=code.generators_x
    stabilizers_z=code.generators_z

    ##  获取逻辑算符
    logical_x=code.logical_operators_x
    logical_z=code.logical_operators_z

    ##  获取数目
    stabilizer_number = len(stabilizers_x) + len(stabilizers_z)
    data_number=code.physical_number
    pauli_number=data_number+stabilizer_number

    ##  ---生成线路---
    ##  强制初始化
    circuit_x = Circuit()
    circuit_z = Circuit()
    circuit_x.append({'name':'R','target':range(data_number)})
    circuit_z.append({'name':'R','target':range(data_number)})

    ##  第一轮测量假设完美的初始化
    observable_include = []  # 记录可观测量的索引
    for i in range(len(logical_x)):
        circuit_x.append({"name": "MPP", "target": logical_x[i]})
        circuit_z.append({"name": "MPP", "target": logical_z[i]})
        observable_include.append(len(circuit_x._measurements)-1)
    for i, stabilizer in enumerate(stabilizers_x):
        circuit_x.append({"name": "MPP", "target": stabilizer})
        circuit_z.append({"name": "MPP", "target": stabilizer})
    for i, stabilizer in enumerate(stabilizers_z):
        circuit_z.append({"name": "MPP", "target": stabilizer})
        circuit_x.append({"name": "MPP", "target": stabilizer})

    ##  施加第一轮噪声
    for i in range(pauli_number):
        circuit_z.append({"name": "DEPOLARIZE1", "target": i, "p": p_noise})

    ##  循环多轮，多轮测量错误与量子噪声信道
    for _ in range(cycle_number):

        ##  测量稳定子
        for i, stabilizer in enumerate(stabilizers_x):
            sequence_temp = syndrome_pauli_css_measurement_circuit(stabilizer, i + data_number, 'x', p_noise, p_measure)
            for temp in sequence_temp:
                circuit_z.append(temp)
                circuit_x.append(temp)
        for i, stabilizer in enumerate(stabilizers_z):
            sequence_temp = syndrome_pauli_css_measurement_circuit(stabilizer, i + data_number + len(stabilizers_x), 'z', p_noise, p_measure)
            for temp in sequence_temp:
                circuit_z.append(temp)
                circuit_x.append(temp)

        ##  添加检测器
        for i in range(stabilizer_number):
            circuit_z.append({"name": "DETECTOR", "target": [-i - 1, -i - stabilizer_number - 1]})
            circuit_x.append({"name": "DETECTOR", "target": [-i - 1, -i - stabilizer_number - 1]})

    ##  最后一轮测量假设是没有噪声的
    ##  测量稳定子
    for i, stabilizer in enumerate(stabilizers_x):
        circuit_z.append({"name": "MPP", "target": stabilizer})
        circuit_x.append({"name": "MPP", "target": stabilizer})
    for i, stabilizer in enumerate(stabilizers_z):
        circuit_z.append({"name": "MPP", "target": stabilizer})
        circuit_x.append({"name": "MPP", "target": stabilizer})

    ##  添加检测器
    for i in range(stabilizer_number):
        circuit_z.append({"name": "DETECTOR", "target": [-i - 1, -i - stabilizer_number - 1]})
        circuit_x.append({"name": "DETECTOR", "target": [-i - 1, -i - stabilizer_number - 1]})

    ##  测量逻辑算符
    for i in range(len(logical_x)):
        circuit_z.append({"name": "MPP", "target": logical_z[i]})
        circuit_x.append({"name": "MPP", "target": logical_x[i]})
        circuit_z.append({"name": "OBSERVABLE_INCLUDE", "target": [len(circuit_z._measurements)-1, observable_include[i]]})
        circuit_x.append({"name": "OBSERVABLE_INCLUDE", "target": [len(circuit_x._measurements)-1, observable_include[i]]})

    ##  ---返回线路---
    return circuit_x, circuit_z


#%%  TODO：===将Majorana CSS code转换为码能力下的测试线路===
def MajoranaCSSCode2CodeCapacityCircuit(code,p_noise,p_measure,cycle_number:int):
    pass


#%%  TODO：===将Pauli CSS code转换为码能力下的测试线路===
def PauliCSSCode2CodeCapacityCircuit(code,p_noise,p_measure,cycle_number:int):
    pass


# %%  KEY：===生成Majorana CSS stabilizer测量线路===
def syndrome_majorana_css_measurement_circuit(stabilizer, qubit_index, type, p_noise, p_measure):

    ##  ———数据预处理———
    sequence = []  # 线路序列
    flag = True  # 门类型标志

    ##  将qubit置于负号匹配
    sequence.append({'name': 'X', 'target': qubit_index})
    sequence.append({'name': 'DEPOLARIZE1', 'target': qubit_index, 'p': p_noise})

    ##  判断稳定子类型
    if type == 'x' or type == 'X':
        occupy=stabilizer.occupy_x
    elif type == 'z' or type == 'Z':
        occupy=stabilizer.occupy_z
    else:
        raise ValueError

    ##  ---生成纠缠线路---
    ##  生成前一半线路
    for j in range(len(occupy)):
        majorana_index_now = occupy[j]

        ##  最后一位与qubit作用CNX gate
        if j == len(occupy) - 1:
            sequence.append({'name': 'CNX', 'target': [majorana_index_now, qubit_index], })
            sequence.append({'name': 'FDEPOLARIZE1', 'target': majorana_index_now, 'p': p_noise})
            sequence.append({'name': 'DEPOLARIZE1', 'target': qubit_index, 'p': p_noise})
            break

        majorana_index_down = occupy[j + 1]  # 后一个fermionic site
        ##  作用braid gate
        if flag:
            ##  根据稳定子类型选择braid形式
            if type == 'X' or type == 'x':
                order_target = [majorana_index_down, majorana_index_now]
            elif type == 'Z' or type == 'z':
                order_target = [majorana_index_now, majorana_index_down]
            else:
                raise ValueError

            ##  添加braid gate
            sequence.append({"name": "BRAID", "target": order_target, })
            sequence.append({'name': 'FDEPOLARIZE1', 'target': order_target, 'p': p_noise})
            flag = False

        ##  作用CNN gate
        else:
            sequence.append({'name': 'CNN', 'target': [majorana_index_now, majorana_index_down], })
            sequence.append({'name': 'FDEPOLARIZE1', 'target': [majorana_index_now, majorana_index_down], 'p': p_noise})
            flag = True

    ##  生成syndrome extraction circuit的另一半
    flag = True
    for j in range(len(occupy) - 1):
        majorana_index_now = occupy[-1 - j]  # 当前的fermionic site
        majorana_index_up = occupy[-1 - j - 1]  # 上一个fermionic site

        ##  作用braid gate
        if flag:
            if type == 'X' or type == 'x':
                order_target = [majorana_index_now, majorana_index_up]
            elif type == 'Z' or type == 'z':
                order_target = [majorana_index_up, majorana_index_now]
            else:
                raise ValueError
            sequence.append({'name': 'N', 'target': [majorana_index_now]})
            sequence.append({'name': 'FDEPOLARIZE1', 'target': majorana_index_now, 'p': p_noise})
            sequence.append({'name': 'BRAID', 'target': order_target})
            sequence.append({'name': 'FDEPOLARIZE1', 'target': order_target, 'p': p_noise})
            sequence.append({'name': 'N', 'target': [majorana_index_now]})
            sequence.append({'name': 'FDEPOLARIZE1', 'target': majorana_index_now, 'p': p_noise})
            flag = False

        ##  作用CNN gate
        else:
            sequence.append({'name': 'CNN', 'target': [majorana_index_now, majorana_index_up]})
            sequence.append({'name': 'FDEPOLARIZE1', 'target': [majorana_index_now, majorana_index_up], 'p': p_noise})
            flag = True

    ##  在qubit上测量结果并重置
    sequence.append({'name': 'MZ', 'target': qubit_index, 'p': p_measure})
    sequence.append({'name': 'R', 'target': qubit_index})
    sequence.append({'name': 'DEPOLARIZE1', 'target': qubit_index, 'p': p_noise})

    ##  ———返回线路序列———
    return sequence


#%%  KEY：===生成Pauli CSS stabilizer测量线路===
def syndrome_pauli_css_measurement_circuit(stabilizer, qubit_index, type, p_noise, p_measure):

    ##  ———数据预处理———
    sequence = []  # 线路序列

    ##  判断稳定子类型
    if type == 'x' or type == 'X':
        occupy=stabilizer.occupy_x
    elif type == 'z' or type == 'Z':
        occupy=stabilizer.occupy_z
    else:
        raise ValueError

    ##  ---生成纠缠线路---
    for j in range(len(occupy)):
        if type == 'Z' or type == 'z':
            sequence.append({'name': 'CX', 'target': [occupy[j], qubit_index]})
            sequence.append({'name': 'DEPOLARIZE1', 'target': [occupy[j], qubit_index], 'p': p_noise})
        elif type == 'X' or type == 'x':
            sequence.append({'name': 'H', 'target': qubit_index})
            sequence.append({'name': 'DEPOLARIZE1', 'target': qubit_index, 'p': p_noise})
            sequence.append({'name': 'CX', 'target': [qubit_index,occupy[j]]})
            sequence.append({'name': 'DEPOLARIZE1', 'target': [occupy[j],qubit_index], 'p': p_noise})
            sequence.append({'name': 'H', 'target': qubit_index})
            sequence.append({'name': 'DEPOLARIZE1', 'target': qubit_index, 'p': p_noise})
        else:
            raise ValueError

    ##  在qubit上测量结果并重置
    sequence.append({'name': 'MZ', 'target': qubit_index, 'p': p_measure})
    sequence.append({'name': 'R', 'target': qubit_index})
    sequence.append({'name': 'DEPOLARIZE1', 'target': qubit_index, 'p': p_noise})

    ##  ———返回线路序列———
    return sequence



"""""
模块作用：定义量子电路的抽象结构，支持指令追加、执行、采样与转换为Stim电路。
"""""
import copy
from multiprocessing import Pool
import pymatching
import qiskit
import numpy as np
import stim
import stimbposd
from qiskit.circuit import CircuitError
from qiskit.circuit.library import XGate, ZGate
from extendedstim.Physics.MajoranaOperator import MajoranaOperator
from extendedstim.Physics.PauliOperator import PauliOperator
from extendedstim.Platform.Frame import Frame
from extendedstim.Platform.Platform import Platform
from extendedstim.tools.TypingTools import isinteger, islist


#%%  KEY：===尝试导入Tesseract类===
try:
    from tesseract_decoder import tesseract
except ImportError:
    print('当前环境不支持tesseract_decoder')
    class tesseract:
        def __init__(self):
            pass

        @staticmethod
        def TesseractConfig(dem, det_beam=50):
            return Config(dem, det_beam)

    class Config:
        def __init__(self, dem: stim.DetectorErrorModel, det_beam):
            self.dem=dem
            self.det_beam=det_beam
        def compile_decoder(self):
            return stimbposd.bp_osd.BPOSD(model=self.dem, bp_method='min_sum', max_bp_iters=self.det_beam)


class Circuit:

    #%%  USER：===构造方法===
    def __init__(self):
        self.majorana_number=0  # fermionic sites的数目
        self.pauli_number=0  # qubits的数目
        self.sequence=[]  # 量子线路的操作序列
        self.noise=[]  # 量子线路中噪声的索引
        self._sequence:list[dict]=[]  # 量子线路的操作序列（真实计算使用）
        self._measurements=[]  # 量子线路中测量的索引
        self._detectors=[]  # 量子线路中测量的结果探测器对self.measurements的索引
        self._observables=[]  # 量子线路中可观测量的对self.measurements的索引
        self._dem=None  # 量子线路错误模型
        self._dem_str_list=[]  # 量子线路错误模型的字符串表示
        self._reference_circuit=None
        self._reference_measurement_sample=None  # 参考样本的测量值
        self._reference_detector_sample=None  # 参考样本的检测值
        self._reference_observable_sample=None  # 参考样本的可观测量值

    #%%  USER：===重载运算符===
    ##  USER：---获取序列中的元素---
    def __getitem__(self, item: int) -> dict:
        return self.sequence[item]

    #%%  USER：===属性方法===
    ##  KEY：---生成无噪声的线路---
    @property
    def reference_circuit(self):
        if self._reference_circuit is not None:
            return self._reference_circuit
        else:
            sequence=copy.deepcopy(self._sequence)  # 深拷贝操作序列
            for i in range(len(self.noise)):
                gate=sequence[self.noise[i]]
                assert isinstance(gate, dict)
                gate['p']=0
            ideal_circuit=copy.deepcopy(self)
            ideal_circuit._sequence=sequence
            return ideal_circuit

    ##  USER：获取参考样本
    @property
    def reference_sample(self):
        if self._reference_measurement_sample is None:
            reference_measurement_sample, reference_detector_sample, reference_observable_sample=self.reference_circuit.execute()
            self._reference_measurement_sample=reference_measurement_sample
            self._reference_detector_sample=reference_detector_sample
            self._reference_observable_sample=reference_observable_sample
            return reference_measurement_sample, reference_detector_sample, reference_observable_sample
        else:
            return self._reference_measurement_sample, self._reference_detector_sample, self._reference_observable_sample

    ##  USER：--生成检测错误模型--
    def detector_error_model(self) -> stim.DetectorErrorModel:
        if self._dem is not None:
            return self._dem
        measurement_sample_origin, detector_sample_origin, observable_sample_origin=self.reference_sample

        ##  执行检验线路的稳定性
        for time in range(5):
            measurement_sample, detector_sample, observable_sample=self.noiseless_sample()
            assert np.all(detector_sample==detector_sample_origin), f'原始线路的detector不是稳定的'
            assert np.all(observable_sample==observable_sample_origin), f'原始线路的observable不是稳定的'

        errors=[]
        dem_str=''
        sample_number=len(self.noise)//20
        noise_group=[self.noise[temp*sample_number:(temp+1)*sample_number] for temp in range(20)]
        if len(self.noise[20*sample_number:])>0:
            noise_group.append(self.noise[20*sample_number:])

        with Pool(processes=len(noise_group)) as pool:
            results=[pool.apply_async(self.noise_sample_batch,args=(noise_group[i],)) for i in range(len(noise_group))]
            final_results=[result.get() for result in results]
        flag=0
        for i in range(len(final_results)):
            for j in range(len(final_results[i])):
                ##  计算错误位置
                measurement_sample, detector_sample, observable_sample=final_results[i][j]
                detector_sample_diff=[detector_sample_origin[j]^detector_sample[j] for j in range(len(detector_sample))]
                observable_sample_diff=[observable_sample_origin[j]^observable_sample[j] for j in range(len(observable_sample))]
                errors.append(len(errors))
                detectors_trigger=np.where(np.array(detector_sample_diff)==True)[0]
                observables_trigger=np.where(np.array(observable_sample_diff)==True)[0]

                ##  合成错误语句
                if len(detectors_trigger)>0 or len(observables_trigger)>0:
                    temp_error=f'error({self._sequence[self.noise[flag]]['p']}) '
                    temp_trigger=''
                    for index in detectors_trigger:
                        temp_trigger=temp_trigger+f' D{index}'
                    for index in observables_trigger:
                        temp_trigger=temp_trigger+f' L{index}'
                    temp=temp_error+temp_trigger
                    if self._sequence[self.noise[flag]]['name']=='M_ERROR':
                        self._dem_str_list.append([temp_error, temp_trigger, 'M_ERROR'])
                    else:
                        self._dem_str_list.append([temp_error, temp_trigger, 'G_ERROR'])
                    dem_str+=('\n'+temp)
                    flag+=1

        ##  返回错误模型
        dem=stim.DetectorErrorModel(dem_str)
        self._dem=dem
        return dem

    #%%  USER：===对象方法===
    ##  USER：---添加量子线路组分操作---
    def append(self, params):
        """""
        {
        'name': str，线路操作的名称
        'target': list or int，操作作用的对象
        'p': float，操作对应的概率，不一定有
        'index': int，测量对应平台上的stabilizers的序号
        'majorana_state': list of MajoranaOperator，强制初始化的majorana state
        'pauli_state': list of PauliOperator，强制初始化的pauli state
        'pauli_number': int，强制初始化的qubits数目
        'majorana_number': int，强制初始化的fermionic sites数目
        }
        支持的线路操作名称：
        'X', 'Y', 'Z', 'H', 'S'：{'name':str,'target':int or list}, single qubit上的qugate
        'X_ERROR', 'Y_ERROR', 'Z_ERROR', 'DEPOLARIZE': {'name':str,'target':int or list}, single qubit上的噪声
        'U', 'V', 'N', 'P'：{'name':str,'target':int or list}, single fermionic site上的fgate
        'U_ERROR', 'V_ERROR', 'N_ERROR', 'FDEPOLARIZE1': {'name':str,'target':int or list}, single fermionic site上的噪声
        'CX', 'CNX', 'CNN', 'BRAID': {'name':str,'target':list}, two qubit or fermionic sites上的gates
        'R','FR': {'name':str,'target':int or list}, single qubit or single fermionic sites上的重置到空态或0态
        'MZ', 'MN': {'name':str,'target':int or list}, single qubit or single fermionic sites上的measurement
        'MPP': {'name':str,'target':list or Operator}, Pauli string operators or Majorana string operators的measurement
        'FORCE'：{'name':str,'pauli_number':int,'majorana_number':int,'pauli_state':list of PauliOperator,'majorana_state':list of MajoranaOperator}, 强制初始化
        'DETECTOR': {'name':str,'target':list of int}，探测器
        'OBSERVABLE_INCLUDE': {'name':str,'target':list of int}，可观测量
        """""

        ##  ---数据预处理---
        assert isinstance(params, dict)
        assert 'name' in params
        name=params["name"]  # 线路操作的名称

        ##  ---添加量子线路操作---
        ##  添加single gate
        if name in ['X', 'Y', 'Z', 'H', 'S', 'X_ERROR', 'Y_ERROR', 'Z_ERROR', 'U', 'V', 'N', 'P', 'U_ERROR', 'V_ERROR', 'N_ERROR']:

            ##  作用在整数index对应目标上
            if isinteger(params['target']):
                self._sequence.append(params.copy())
                self.sequence.append(self._sequence[-1])

            ##  作用在一系列目标上
            elif islist(params['target']):
                for temp in params['target']:
                    params_temp={'name': name, 'target': temp}
                    self.append(params_temp)

        ##  添加two gate
        elif name in ['CX', 'CNX', 'CNN']:

            ##  添加单个two gate
            if islist(params['target']) and len(params['target'])==2 and isinteger(params['target'][0]) and isinteger(params['target'][1]):
                self._sequence.append(params.copy())
                self.sequence.append(self._sequence[-1])

            ##  用单个列表添加多个two gates
            elif islist(params['target']) and isinteger(params['target'][0]):
                for i in range(len(params['target'])//2):
                    params_temp={'name': name, 'target': [params['target'][2*i], params['target'][2*i+1]]}
                    self.append(params_temp)

            ##  用多个列表添加多个two gates
            elif islist(params['target']):
                for temp in params['target']:
                    self.append({'name': name, 'target': temp})

            ##  其他情况抛出异常
            else:
                raise ValueError("CX, CNX, CNN gate must be applied to two")

        ##  添加braid gate
        elif name=='BRAID' or name=='braid':

            if islist(params['target']) and len(params['target'])==2 and isinteger(params['target'][0]) and isinteger(params['target'][1]):
                x=params['target'][0]  # 控制位
                y=params['target'][1]  # 目标位

                ##  控制位等于目标位时，根据verse参数判断是否添加N门
                if x==y and 'verse' not in params:
                    self._sequence.append({'name': 'P', 'target': x})
                    self.sequence.append(self._sequence[-1])
                elif x==y and 'verse' in params and params['verse']==True:
                    self._sequence.append({'name': 'P', 'target': y})
                    self.sequence.append(self._sequence[-1])
                    self._sequence.append({'name': 'N', 'target': y})
                    self.sequence.append(self._sequence[-1])

                ##  一般情况下braid gate作用在两个不同的fermionic sites上
                elif x!=y:
                    self._sequence.append({'name': 'BRAID', 'target': [x, y]})
                    self.sequence.append(self._sequence[-1])

            ##  用单个列表添加多个two gates
            elif islist(params['target']) and isinteger(params['target'][0]):
                for i in range(len(params['target'])//2):
                    params_temp={'name': name, 'target': [params['target'][2*i], params['target'][2*i+1]]}
                    self.append(params_temp)

            ##  用多个列表添加多个two gates
            elif islist(params['target']):
                for temp in params['target']:
                    self.append({'name': name, 'target': temp})

            ##  其他情况抛出异常
            else:
                raise ValueError("CX, CNX, CNN gate must be applied to two")

        ##  添加qubit上的去极化噪声
        elif name=='DEPOLARIZE1':
            assert 'p' in params

            ##  作用在整数index对应目标上
            if isinteger(params['target']):
                self.sequence.append({'name': 'DEPOLARIZE1', 'target': params['target'], 'p': params['p']})
                fix=(1-np.sqrt(1-4*params["p"]/3))/2
                self._sequence.append({'name': 'X_ERROR', 'target': params["target"], 'p': fix})
                self.noise.append(len(self._sequence)-1)
                self._sequence.append({'name': 'Y_ERROR', 'target': params["target"], 'p': fix})
                self.noise.append(len(self._sequence)-1)
                self._sequence.append({'name': 'Z_ERROR', 'target': params["target"], 'p': fix})
                self.noise.append(len(self._sequence)-1)

            ##  作用在一系列目标上
            elif islist(params['target']):
                for temp in params['target']:
                    self.append({'name': params['name'], 'target': temp, 'p': params["p"]})

        ##  添加fermionic site上的去极化噪声
        elif name=='FDEPOLARIZE1':
            assert 'p' in params

            ##  作用在整数index对应目标上
            if isinteger(params['target']):
                self.sequence.append({'name': 'FDEPOLARIZE1', 'target': params['target'], 'p': params['p']})
                fix=(1-np.sqrt(1-4*params["p"]/3))/2
                self._sequence.append({'name': 'U_ERROR', 'target': params["target"], 'p': fix})
                self.noise.append(len(self._sequence)-1)
                self._sequence.append({'name': 'V_ERROR', 'target': params["target"], 'p': fix})
                self.noise.append(len(self._sequence)-1)
                self._sequence.append({'name': 'N_ERROR', 'target': params["target"], 'p': fix})
                self.noise.append(len(self._sequence)-1)

            ##  作用在一系列目标上
            elif islist(params['target']):
                for temp in params['target']:
                    self.append({'name': params['name'], 'target': temp, 'p': params["p"]})

        ##  添加single qubit and single fermionic site上的测量
        elif name=='MZ' or name=='MN':
            target=params['target']

            ##  作用在一系列目标上
            if islist(target):
                for temp in target:
                    dict_temp={'name': name, 'target': temp}
                    self.append(dict_temp)

            ##  作用在整数index对应目标上
            elif isinteger(target):
                if name=='MZ':
                    dict_temp={'name': 'MPP', 'target': PauliOperator([], [target], 1)}
                else:
                    dict_temp={'name': 'MPP', 'target': MajoranaOperator([target], [target], 1j)}
                self.append(dict_temp)

        ##  添加qubit重置
        elif name=='R':
            assert 'target' in params
            target=params['target']

            ##  作用在整数index对应目标上
            if isinteger(target):
                self._sequence.append({'name': name, 'target': target})
                self.sequence.append(self._sequence[-1])
                if target==self.pauli_number:
                    self.pauli_number+=1
                elif target>self.pauli_number or target<0:
                    raise ValueError("R gate target must be consecutive")
                else:
                    pass
            elif islist(target):
                for temp in target:
                    self.append({'name': name, 'target': temp})
            else:
                raise ValueError

        ##  添加fermionic site重置
        elif name=='FR':
            assert 'target' in params
            target=params['target']
            if isinteger(target):
                self._sequence.append({'name': name, 'target': target})
                self.sequence.append(self._sequence[-1])
                if target==self.majorana_number:
                    self.majorana_number=target+1
                elif target>self.majorana_number or target<0:
                    raise ValueError("FR gate target must be consecutive")
                else:
                    pass
            elif islist(target):
                for temp in target:
                    self.append({'name': name, 'target': temp})
            else:
                raise ValueError

        ##  强制初始化
        elif name=='FORCE':
            assert 'majorana_state' in params
            assert 'pauli_state' in params
            assert 'pauli_number' in params
            assert 'majorana_number' in params
            stabilizers_majorana: list[MajoranaOperator]=params["majorana_state"]
            stabilizers_pauli: list[PauliOperator]=params["pauli_state"]
            self.pauli_number=params["pauli_number"]
            self.majorana_number=params["majorana_number"]
            for i in range(len(stabilizers_majorana)):
                assert len(stabilizers_majorana[i].occupy_x)==0 or np.max(stabilizers_majorana[i].occupy_x)<self.majorana_number
                assert len(stabilizers_majorana[i].occupy_z)==0 or np.max(stabilizers_majorana[i].occupy_z)<self.majorana_number
            for i in range(len(stabilizers_pauli)):
                assert len(stabilizers_pauli[i].occupy_x)==0 or np.max(stabilizers_pauli[i].occupy_x)<self.pauli_number
                assert len(stabilizers_pauli[i].occupy_z)==0 or np.max(stabilizers_pauli[i].occupy_z)<self.pauli_number

            self._sequence.append({'name': 'FORCE', 'majorana_state': stabilizers_majorana, 'pauli_state': stabilizers_pauli})
            self.sequence.append(self._sequence[-1])

        ##  添加string算符的测量
        elif name=='MPP':
            assert 'target' in params
            if islist(params['target']):
                for i, temp in enumerate(params['target']):
                    dict_temp={'name': 'MPP', 'target': temp}
                    if 'index' in params:
                        dict_temp['index']=params['index'][i]
                    self.append(dict_temp)
            elif isinstance(params['target'], (PauliOperator, MajoranaOperator)):
                dict_temp={'name': 'MPP', 'target': params['target']}
                if 'index' in params:
                    dict_temp['index']=params['index']
                self._sequence.append(dict_temp)
                self._measurements.append(len(self._sequence)-1)
                if 'p' in params:
                    self._sequence.append({'name': 'M_ERROR', 'p': params["p"]})
                    self.noise.append(len(self._sequence)-1)
                    self.sequence.append({'name': 'MPP', 'target': params['target'], 'p': params["p"]})
                else:
                    self.sequence.append(self._sequence[-1])
            else:
                raise ValueError

        ##  添加监视器
        elif name=='DETECTOR':
            assert 'target' in params
            target=params['target']
            if all(target[i]<0 for i in range(len(target))):
                together=[len(self._measurements)+temp for temp in target]  # 在测量中找到对应索引
                self._detectors.append(together)
                self.sequence.append({'name': 'DETECTOR', 'target': target})
            elif all(target[i]>=0 for i in range(len(target))):
                self._detectors.append([temp for temp in target])
                self.sequence.append({'name': 'DETECTOR', 'target': [-len(self._measurements)+temp for temp in target]})
            else:
                raise ValueError("DETECTOR gate target must be consecutive")

        ##  添加可观测量
        elif name=='OBSERVABLE_INCLUDE':
            assert 'target' in params
            target=params['target']
            if all(target[i]<0 for i in range(len(target))):
                together=[len(self._measurements)+temp for temp in target]  # 在测量中找到索引
                self._observables.append(together)
                self.sequence.append({'name': 'OBSERVABLE_INCLUDE', 'target': target})
            elif all(target[i]>=0 for i in range(len(target))):
                together=[temp for temp in target]  # 在测量中找到索引
                self._observables.append(together)
                self.sequence.append({'name': 'OBSERVABLE_INCLUDE', 'target': [-len(self._measurements)+temp for temp in target]})
            else:
                raise ValueError("OBSERVABLE_INCLUDE gate target must be consecutive")
        else:
            raise NotImplementedError

    ##  USER：---执行线路并返回测量结果---
    def execute(self):

        ##  ---生成量子平台---
        platform=Platform()  # 生成量子平台
        platform.initialize(self.majorana_number, self.pauli_number)  # 初始化一定数目的qubits和fermionic sites
        measurement_sample=np.empty(len(self._measurements), dtype=int)  # 生成测量值的样本数组
        flag_measurement=0

        ##  遍历整个操作序列
        for i, gate in enumerate(self._sequence):
            name=gate['name']

            ##  执行单门
            if name in ['X', 'Y', 'Z', 'H', 'S', 'U', 'V', 'N']:
                target: int=gate['target']
                if name=='X':
                    platform.x(target)
                elif name=='Y':
                    platform.y(target)
                elif name=='Z':
                    platform.z(target)
                elif name=='H':
                    platform.h(target)
                elif name=='S':
                    platform.s(target)
                elif name=='U':
                    platform.u(target)
                elif name=='V':
                    platform.v(target)
                elif name=='N':
                    platform.n(target)

            ##  执行双门
            elif name in ['CX', 'CNX', 'BRAID', 'CNN']:
                target: list=gate['target']
                if name=='CX':
                    platform.cx(target[0], target[1])
                elif name=='CNX':
                    platform.cnx(target[0], target[1])
                elif name=='BRAID':
                    platform.braid(target[0], target[1])
                elif name=='CNN':
                    platform.cnn(target[0], target[1])

            ##  执行重置
            elif name=='R':
                target: int=gate['target']
                platform.reset(target)

            elif name=='FR':
                target: int=gate['target']
                platform.fermionic_reset(target)

            ##  执行误差门
            elif name in ['X_ERROR', 'Y_ERROR', 'Z_ERROR', 'U_ERROR', 'V_ERROR', 'N_ERROR']:
                target: int=gate['target']
                p: float=gate['p']
                if name=='X_ERROR':
                    platform.x_error(target, p)
                elif name=='Y_ERROR':
                    platform.y_error(target, p)
                elif name=='Z_ERROR':
                    platform.z_error(target, p)
                elif name=='U_ERROR':
                    platform.u_error(target, p)
                elif name=='V_ERROR':
                    platform.v_error(target, p)
                elif name=='N_ERROR':
                    platform.n_error(target, p)

            elif name=='M_ERROR':
                p: float=gate['p']
                if np.random.rand()<p:
                    measurement_sample[flag_measurement-1]=-measurement_sample[flag_measurement-1]

            ##  执行测量
            elif name=='MPP':
                target=gate['target']
                measurement_sample[flag_measurement]=platform.measure(target)
                flag_measurement+=1

            ##  执行强制初始化
            elif name=='FORCE':
                assert gate['majorana_state'] is not None
                assert gate['pauli_state'] is not None
                platform.force(gate['majorana_state'], gate['pauli_state'])

            ##  其他类型抛出错误
            else:
                raise ValueError(f"Gate {name} is illegal")

        ##  ---返回可观测的结果---
        detector_sample=diff(measurement_sample, self._detectors)
        observable_sample=diff(measurement_sample, self._observables)
        return measurement_sample, detector_sample, observable_sample

    ##  USER：--修改error model的错误几率--
    def noise_amplitude_fix(self, p_noise, p_measure):
        dem_str=''
        for i, temp in enumerate(self._dem_str_list):
            if temp[2]=='G_ERROR':
                temp[0]=f'error({p_noise/3}) '
            elif temp[2]=='M_ERROR':
                temp[0]=f'error({p_measure}) '
            dem_str+=('\n'+temp[0]+temp[1])
            self._dem=stim.DetectorErrorModel(dem_str)

    ##  USER：--生成解码函数--
    def decoder(self, method):
        dem=self.detector_error_model()
        if method=='bposd':
            decoder=stimbposd.bp_osd.BPOSD(model=dem, bp_method='min_sum', max_bp_iters=100)
        elif method=='tesseract':
            config=tesseract.TesseractConfig(dem=dem, det_beam=50)
            decoder=config.compile_decoder()
        elif method=='matching':
            decoder=pymatching.Matching()
            decoder.from_detector_error_model(dem)
        else:
            raise NotImplementedError
        return decoder

    ##  USER：---获取采样---
    def sample(self):
        frame=Frame()
        frame.initialize(self.majorana_number, self.pauli_number)
        reference_measurement_sample, reference_detector_sample, reference_observable_sample=self.reference_sample
        measurement_sample=np.empty(len(self._measurements), dtype=int)  # 生成测量值的样本数组
        flag_measurement=0
        ##  遍历整个操作序列
        for i, gate in enumerate(self._sequence):
            name=gate['name']

            ##  执行单门
            if name in ['X', 'Y', 'Z', 'H', 'S', 'U', 'V', 'N']:
                pass

            ##  执行双门
            elif name in ['CX', 'CNX', 'BRAID', 'CNN']:
                target: list=gate['target']
                if name=='CX':
                    frame.cx(target[0], target[1])
                elif name=='CNX':
                    frame.cnx(target[0], target[1])
                elif name=='BRAID':
                    frame.braid(target[0], target[1])
                elif name=='CNN':
                    frame.cnn(target[0], target[1])

            ##  执行重置
            elif name=='R':
                target: int=gate['target']
                frame.reset(target)

            elif name=='FR':
                target: int=gate['target']
                frame.fermionic_reset(target)

            ##  执行误差门
            elif name in ['X_ERROR', 'Y_ERROR', 'Z_ERROR', 'U_ERROR', 'V_ERROR', 'N_ERROR']:
                target: int=gate['target']
                p: float=gate['p']
                if name=='X_ERROR':
                    frame.x_error(target, p)
                elif name=='Y_ERROR':
                    frame.y_error(target, p)
                elif name=='Z_ERROR':
                    frame.z_error(target, p)
                elif name=='U_ERROR':
                    frame.u_error(target, p)
                elif name=='V_ERROR':
                    frame.v_error(target, p)
                elif name=='N_ERROR':
                    frame.n_error(target, p)

            elif name=='M_ERROR':
                p: float=gate['p']
                if np.random.rand()<p:
                    measurement_sample[flag_measurement-1]=-measurement_sample[flag_measurement-1]

            ##  执行测量
            elif name=='MPP':
                target=gate['target']
                measurement_sample[flag_measurement]=frame.measure(target, reference_measurement_sample[flag_measurement])
                flag_measurement+=1

            ##  执行强制初始化
            elif name=='FORCE':
                assert gate['majorana_state'] is not None
                assert gate['pauli_state'] is not None
                frame.force(gate['majorana_state'], gate['pauli_state'])

            ##  其他类型抛出错误
            else:
                raise ValueError(f"Gate {name} is illegal")

        ##  ---返回可观测的结果---
        detector_sample=diff(measurement_sample, self._detectors)
        observable_sample=diff(measurement_sample, self._observables)
        return measurement_sample, detector_sample, observable_sample

    def noise_sample_batch(self, noise_index_list):
        results=[self.noise_sample(noise_index) for noise_index in noise_index_list]
        return results

    ##  USER：---获取采样---
    def noiseless_sample(self):
        frame=Frame()
        frame.initialize(self.majorana_number, self.pauli_number)
        reference_measurement_sample, reference_detector_sample, reference_observable_sample=self.reference_sample
        measurement_sample=np.empty(len(self._measurements), dtype=int)  # 生成测量值的样本数组
        flag_measurement=0
        ##  遍历整个操作序列
        for i, gate in enumerate(self._sequence):
            name=gate['name']

            ##  执行单门
            if name in ['X', 'Y', 'Z', 'U', 'V', 'N']:
                pass
            elif name=='H':
                target: int=gate['target']
                frame.h(target)
            elif name=='S':
                target: int=gate['target']
                frame.s(target)
            elif name=='P':
                target: int=gate['target']
                frame.p(target)
            ##  执行双门
            elif name in ['CX', 'CNX', 'BRAID', 'CNN']:
                target: list=gate['target']
                if name=='CX':
                    frame.cx(target[0], target[1])
                elif name=='CNX':
                    frame.cnx(target[0], target[1])
                elif name=='BRAID':
                    frame.braid(target[0], target[1])
                elif name=='CNN':
                    frame.cnn(target[0], target[1])

            ##  执行重置
            elif name=='R':
                target: int=gate['target']
                frame.reset(target)

            elif name=='FR':
                target: int=gate['target']
                frame.fermionic_reset(target)

            ##  执行误差门
            elif name in ['X_ERROR', 'Y_ERROR', 'Z_ERROR', 'U_ERROR', 'V_ERROR', 'N_ERROR','M_ERROR']:
                pass

            ##  执行测量
            elif name=='MPP':
                target=gate['target']
                measurement_sample[flag_measurement]=frame.measure(target, reference_measurement_sample[flag_measurement])
                flag_measurement+=1

            ##  执行强制初始化
            elif name=='FORCE':
                assert gate['majorana_state'] is not None
                assert gate['pauli_state'] is not None
                frame.force(gate['majorana_state'], gate['pauli_state'])

            ##  其他类型抛出错误
            else:
                raise ValueError(f"Gate {name} is illegal")

        ##  ---返回可观测的结果---
        detector_sample=diff(measurement_sample, self._detectors)
        observable_sample=diff(measurement_sample, self._observables)
        return measurement_sample, detector_sample, observable_sample

    ##  USER：---获取采样---
    def noise_sample(self,noise_index):
        frame=Frame()
        frame.initialize(self.majorana_number, self.pauli_number)
        reference_measurement_sample, reference_detector_sample, reference_observable_sample=self.reference_sample
        measurement_sample=np.empty(len(self._measurements), dtype=int)  # 生成测量值的样本数组
        flag_measurement=0
        ##  遍历整个操作序列
        for i, gate in enumerate(self._sequence):
            name=gate['name']

            ##  执行单门
            if name in ['X', 'Y', 'Z','U', 'V', 'N']:
                pass
            elif name=='H':
                target: int=gate['target']
                frame.h(target)
            elif name=='S':
                target: int=gate['target']
                frame.s(target)
            elif name=='P':
                target: int=gate['target']
                frame.p(target)

            ##  执行双门
            elif name in ['CX', 'CNX', 'BRAID', 'CNN']:
                target: list=gate['target']
                if name=='CX':
                    frame.cx(target[0], target[1])
                elif name=='CNX':
                    frame.cnx(target[0], target[1])
                elif name=='BRAID':
                    frame.braid(target[0], target[1])
                elif name=='CNN':
                    frame.cnn(target[0], target[1])

            ##  执行重置
            elif name=='R':
                target: int=gate['target']
                frame.reset(target)

            elif name=='FR':
                target: int=gate['target']
                frame.fermionic_reset(target)

            ##  执行误差门
            elif name in ['X_ERROR', 'Y_ERROR', 'Z_ERROR', 'U_ERROR', 'V_ERROR', 'N_ERROR']:
                if noise_index==i:
                    target: int=gate['target']
                    if name=='X_ERROR':
                        frame.x_error(target, 1)
                    elif name=='Y_ERROR':
                        frame.y_error(target, 1)
                    elif name=='Z_ERROR':
                        frame.z_error(target, 1)
                    elif name=='U_ERROR':
                        frame.u_error(target, 1)
                    elif name=='V_ERROR':
                        frame.v_error(target, 1)
                    elif name=='N_ERROR':
                        frame.n_error(target, 1)

            elif name=='M_ERROR':
                if noise_index==i:
                    measurement_sample[flag_measurement-1]=-measurement_sample[flag_measurement-1]

            ##  执行测量
            elif name=='MPP':
                target=gate['target']
                measurement_sample[flag_measurement]=frame.measure(target, reference_measurement_sample[flag_measurement])
                flag_measurement+=1

            ##  执行强制初始化
            elif name=='FORCE':
                assert gate['majorana_state'] is not None
                assert gate['pauli_state'] is not None
                frame.force(gate['majorana_state'], gate['pauli_state'])

            ##  其他类型抛出错误
            else:
                raise ValueError(f"Gate {name} is illegal")

        ##  ---返回可观测的结果---
        detector_sample=diff(measurement_sample, self._detectors)
        observable_sample=diff(measurement_sample, self._observables)
        return measurement_sample, detector_sample, observable_sample

    ##  USER：--执行线路并返回错误率--
    def experiment(self, sample_number: int, method: str):

        ##  生成线路执行样本和每个样本的预测样本
        dem=self.detector_error_model()  # 错误模型
        sampler=dem.compile_sampler()  # 采样器
        decoder=self.decoder(method)  # 解码器
        detector_data, obs_data, error_data=sampler.sample(shots=sample_number)  # 样本
        predictions=decoder.decode_batch(detector_data)  # 解码器对每个样本的预测
        ##  计算逻辑错误率
        num_errors=0
        for shot in range(sample_number):
            actual_for_shot=obs_data[shot]
            predicted_for_shot=predictions[shot]
            if not np.array_equal(actual_for_shot, predicted_for_shot):
                num_errors+=1

        ##  返回结果
        return num_errors/sample_number

    ##  USER：--生成stim的线路--
    def stim_circuit(self):
        circuit=stim.Circuit()
        flag_measure=0
        for i in range(len(self.sequence)):
            gate=self.sequence[i]
            name=gate['name']
            if name=='X' or name=='Y' or name=='Z' or name=='H' or name=='S' or name=='P':
                circuit.append(name, [gate['target']])

            ##  添加single-qubit上的噪声
            elif name=='X_ERROR' or name=='Y_ERROR' or name=='Z_ERROR':
                circuit.append(name, [gate['target']], gate['p'])

            ##  添加single-fermionic-site gate
            elif name in ['U', 'V', 'N', 'P', 'U_ERROR', 'V_ERROR', 'N_ERROR', 'CNX', 'CNN', 'B', 'braid', 'MN', 'FDEPOLARIZE1', 'FR']:
                raise NotImplementedError('stim只支持pauli circuit')

            ##  添加受控非门
            elif name=='CX':
                target=gate['target']
                circuit.append(name, target)

            ##  添加qubit上的去极化噪声
            elif name=='DEPOLARIZE1':
                target=gate['target']
                circuit.append(name, target, gate['p'])

            ##  强制初始化
            elif name=='FORCE':
                raise NotImplementedError('stim不支持强制初始化')

            ##  添加string算符的测量
            elif name=='MPP':

                ##  求string算符格式化表示
                op: PauliOperator=gate['target']
                occupy_x=op.occupy_x
                occupy_z=op.occupy_z

                ##  简单测量
                if len(occupy_x)==0 and len(occupy_z)==1:
                    if 'p' in gate:
                        circuit.append('MZ', [occupy_z[0]], gate['p'])
                    else:
                        circuit.append('MZ', [occupy_z[0]])
                    continue
                elif len(occupy_x)==1 and len(occupy_z)==0:
                    if 'p' in gate:
                        circuit.append('MX', [occupy_x[0]], gate['p'])
                    else:
                        circuit.append('MX', [occupy_x[0]])
                    continue
                elif len(occupy_x)==1 and len(occupy_z)==1 and occupy_x[0]==occupy_z[0]:
                    if 'p' in gate:
                        circuit.append('MY', [occupy_z[0]], gate['p'])
                    else:
                        circuit.append('MY', [occupy_z[0]])
                    continue

                ##  string operator测量
                op_str=''
                for i in range(self.pauli_number):
                    if i in occupy_x and i in occupy_z:
                        op_str+='Y'
                    elif i in occupy_z:
                        op_str+='Z'
                    elif i in occupy_x:
                        op_str+='X'
                    else:
                        op_str+='_'
                if 'p' in gate:
                    circuit.append('MPP', [stim.PauliString(op_str)], gate['p'])
                else:
                    circuit.append('MPP', [stim.PauliString(op_str)])

            ##  添加qubit重置
            elif name=='R':
                circuit.append('R', [gate['target']])

            ##  检测器
            elif name=='DETECTOR':
                circuit.append(name, [stim.target_rec(temp) for temp in gate['target']])

            ##  添加可观测量
            elif name=='OBSERVABLE_INCLUDE':
                circuit.append(name, [stim.target_rec(temp) for temp in gate['target']], flag_measure)
                flag_measure+=1
            else:
                raise NotImplementedError
        return circuit

    ##  USER：--绘制线路图--
    def draw(self, filename):

        # 绘制一个带有barriers和更多寄存器中，绘制一个新的电路
        F=qiskit.QuantumRegister(self.majorana_number, name='F')  # fermionic sites
        Q=qiskit.QuantumRegister(self.pauli_number, name='Q')  # qubits
        C=qiskit.ClassicalRegister(1, name='C')  # 经典寄存器
        A=qiskit.QuantumRegister(1, name='A')  # 无含义的ancilla
        circuit_qiskit=qiskit.QuantumCircuit(F, Q, C, A)
        braid=qiskit.circuit.ControlledGate(name='braid', num_qubits=2, params=[], label=None, num_ctrl_qubits=1, base_gate=XGate())
        cnn=qiskit.circuit.ControlledGate(name='CNN', num_qubits=2, params=[], label=None, num_ctrl_qubits=1, base_gate=ZGate())
        cnx=qiskit.circuit.ControlledGate(name='CNX', num_qubits=2, params=[], label=None, num_ctrl_qubits=1, base_gate=XGate())
        x_error=qiskit.circuit.Gate('X_ERROR', 1, label='X', params=[])
        y_error=qiskit.circuit.Gate('Y_ERROR', 1, label='Y', params=[])
        z_error=qiskit.circuit.Gate('Z_ERROR', 1, label='Z', params=[])
        u_error=qiskit.circuit.Gate('U_ERROR', 1, label='U', params=[])
        v_error=qiskit.circuit.Gate('V_ERROR', 1, label='V', params=[])
        n_error=qiskit.circuit.Gate('N_ERROR', 1, label='N', params=[])
        reset=qiskit.circuit.Gate('reset', 1, label=None, params=[])
        n=qiskit.circuit.Gate('N', 1, label='N', params=[])
        dep=qiskit.circuit.Gate('DEPOLARIZE', 1, label='D', params=[])

        ##  遍历电路序列，添加到qiskit电路中
        for gate in self.sequence:
            if gate['name']=='R':
                circuit_qiskit.append(reset, [Q[gate['target']]])
            elif gate['name']=='FR':
                circuit_qiskit.append(reset, [F[gate['target']]])
            elif gate['name']=='X':
                circuit_qiskit.x(Q[gate['target']])
            elif gate['name']=='Y':
                circuit_qiskit.y(Q[gate['target']])
            elif gate['name']=='Z':
                circuit_qiskit.z(Q[gate['target']])
            elif gate['name']=='H':
                circuit_qiskit.h(Q[gate['target']])
            elif gate['name']=='S':
                circuit_qiskit.s(Q[gate['target']])
            elif gate['name']=='N':
                circuit_qiskit.append(n, [F[gate['target']]])
            elif gate['name']=='CX':
                circuit_qiskit.cx(Q[gate['target'][0]], Q[gate['target'][1]])
            elif gate['name']=='braid':
                circuit_qiskit.append(braid, [F[gate['target'][0]], F[gate['target'][1]]])
            elif gate['name']=='CNN':
                circuit_qiskit.append(cnn, [F[gate['target'][0]], F[gate['target'][1]]])
            elif gate['name']=='CNX':
                circuit_qiskit.append(cnx, [F[gate['target'][0]], Q[gate['target'][1]]])
            elif gate['name']=='X_ERROR':
                circuit_qiskit.append(x_error, [Q[gate['target']]])
            elif gate['name']=='Y_ERROR':
                circuit_qiskit.append(y_error, [Q[gate['target']]])
            elif gate['name']=='Z_ERROR':
                circuit_qiskit.append(z_error, [Q[gate['target']]])
            elif gate['name']=='U_ERROR':
                circuit_qiskit.append(u_error, [F[gate['target']]])
            elif gate['name']=='V_ERROR':
                circuit_qiskit.append(v_error, [F[gate['target']]])
            elif gate['name']=='N_ERROR':
                circuit_qiskit.append(n_error, [F[gate['target']]])
            elif gate['name']=='DEPOLARIZE1':
                circuit_qiskit.append(dep, [Q[gate['target']]])
            elif gate['name']=='FDEPOLARIZE1':
                circuit_qiskit.append(dep, [F[gate['target']]])
            elif gate['name']=='MPP':
                op=gate['target']
                if isinstance(op, MajoranaOperator):
                    f_flag_x=op.occupy_x
                    f_flag_z=op.occupy_z
                    f_flag_n=np.intersect1d(f_flag_x, f_flag_z)
                    f_flag_x=np.setdiff1d(f_flag_x, f_flag_n)
                    f_flag_z=np.setdiff1d(f_flag_z, f_flag_n)
                    f=np.concatenate([f_flag_x, f_flag_z, f_flag_n])
                    if len(f)>1:
                        mppx=qiskit.circuit.ControlledGate(name='MPPX', num_qubits=len(f)+1, params=[], label=None, num_ctrl_qubits=len(f),
                                                           base_gate=XGate())
                        circuit_qiskit.append(mppx, F[f.tolist()]+[A[0]])
                        circuit_qiskit.measure(A[0], C[0])
                    else:
                        circuit_qiskit.measure(F[f[0]], C[0])
                elif isinstance(op, PauliOperator):
                    p_flag_x=op.occupy_x
                    p_flag_z=op.occupy_z
                    p_flag_y=np.intersect1d(p_flag_x, p_flag_z)
                    p_flag_x=np.setdiff1d(p_flag_x, p_flag_y)
                    p_flag_z=np.setdiff1d(p_flag_z, p_flag_y)
                    p=np.concatenate([p_flag_x, p_flag_y, p_flag_z])
                    if len(p)>1:
                        mppx=qiskit.circuit.ControlledGate(name='MPPX', num_qubits=len(p)+1, params=[], label=None, num_ctrl_qubits=len(p),
                                                           base_gate=XGate())
                        circuit_qiskit.append(mppx, Q[p.tolist()]+[A[0]])
                        circuit_qiskit.measure(A[0], C[0])
                    else:
                        circuit_qiskit.measure(Q[p[0]], C[0])
                else:
                    raise CircuitError("cannot set parameters on immutable base gate")
        ##  格式化颜色
        red='#E77081'
        blue='#5375CD'
        green='#00857B'
        grey='#8C92AC'
        purple='#5D548C'
        orange='#F15D22'
        pink='#FFACC5'
        cyan='#C9DCC4'

        ##  绘制线路
        circuit_qiskit.draw(output='mpl', filename=filename, style={
            'displaycolor': {'cx': None, 'cy': None, 'cz': None,
                             'X_ERROR': red, 'Y_ERROR': red, 'Z_ERROR': red,
                             'U_ERROR': red, 'V_ERROR': red, 'N_ERROR': red,
                             'R': cyan, 'measure': grey,
                             'x': blue, 'y': blue, 'z': blue, 's': blue, 'N': blue, 'U': blue, 'V': blue,
                             'CNN': blue, 'CNX': purple, 'braid': pink,
                             'MPPX': grey
                             },
            'fontsize': 12
        })

    ##  USER：---复制函数---
    def copy(self):
        return copy.deepcopy(self)


def diff(measurement_sample,detectors):
    ##  计算探测器的结果
    detector_sample=np.empty(len(detectors), dtype=bool)
    flag_detector=0
    for i, detector in enumerate(detectors):
        value=measurement_sample[detector][0]
        detector_sample[flag_detector]=False
        for temp in measurement_sample[detector]:
            if value==temp:
                continue
            else:
                detector_sample[flag_detector]=True
                break
        flag_detector+=1
    return detector_sample


#%%  USER：===模块导入与别名===
"""
模块作用：集中导入ExtendedStim内部的主要类并提供简化别名，方便用户直接使用。
output.PauliCode 等：分别为各量子编码与电路相关类的别名。
influence：导入时执行底层模块初始化（若有）。不执行任何额外逻辑。
"""
import extendedstim.Code.QuantumCode.QuantumCode
import extendedstim.Code.QuantumCode.QuantumCSSCode
import extendedstim.Code.QuantumCode.PauliCode
import extendedstim.Code.QuantumCode.PauliCSSCode
import extendedstim.Code.QuantumCode.MajoranaCode
import extendedstim.Code.QuantumCode.MajoranaCSSCode
import extendedstim.Code.LinearCode.LinearCode
import extendedstim.Code.LinearCode.BicycleCode
import extendedstim.Code.LinearCode.FiniteEuclideanGeometryCode
import extendedstim.Code.LinearCode.FiniteProjectiveGeometryCode
import extendedstim.Circuit.Circuit
import extendedstim.Circuit.Code2Circuit
import extendedstim.Circuit.Code2Circuit

#%%  USER：===别名导出===
"""
将常用类统一导出为顶层别名，方便用户直接 `from extendedstim import PauliCode` 等调用。
"""
PauliCode=extendedstim.Code.QuantumCode.PauliCode.PauliCode
PauliCSSCode=extendedstim.Code.QuantumCode.PauliCSSCode.PauliCSSCode
MajoranaCode=extendedstim.Code.QuantumCode.MajoranaCode.MajoranaCode
MajoranaCSSCode=extendedstim.Code.QuantumCode.MajoranaCSSCode.MajoranaCSSCode
LinearCode=extendedstim.Code.LinearCode.LinearCode.LinearCode
BicycleCode=extendedstim.Code.LinearCode.BicycleCode.BicycleCode
FiniteEuclideanGeometryCode=extendedstim.Code.LinearCode.FiniteEuclideanGeometryCode.FiniteEuclideanGeometryCode
FiniteProjectiveGeometryCode=extendedstim.Code.LinearCode.FiniteProjectiveGeometryCode.FiniteProjectiveGeometryCode
Code2Circuit=extendedstim.Circuit.Code2Circuit.Code2Circuit
Circuit=extendedstim.Circuit.Circuit.Circuit



"""""
模块作用：提供GF(2)线性代数与组合优化相关的工具方法，如解线性方程、子空间交并、正交化、码距估计等。
"""""
import galois
import numpy as np
from mip import Model, BINARY, minimize, xsum


##  USER：---解GF(2)线性方程组 A^T x = b---
def solve(matrix, vector):
    """""
    input.matrix：galois.GF(2) 或 01矩阵，形状 (m, n)
    input.vector：galois.GF(2) 或 01向量，形状 (n,)
    output：galois.GF(2) 向量解；若无解返回 None
    """""

    ##  ---数据预处理---
    assert matrix.shape[1] == vector.shape[0]
    A = matrix.T
    b = vector.reshape(-1, 1)

    ##  ---求解线性方程---
    aug = np.concatenate((A, b), axis=1)
    n, m_plus_1 = aug.shape
    m = m_plus_1 - 1
    rank = 0
    for col in range(m):
        ##  寻找主元
        pivot_row = None
        for i in range(rank, n):
            if aug[i, col] == 1:
                pivot_row = i
                break
        if pivot_row is None:
            ##  该列没有主元，跳过
            continue
        ##  交换行
        if pivot_row != rank:
            aug[[rank, pivot_row], :] = aug[[pivot_row, rank], :]
        ##  消去当前列下方和上方的元素
        for i in range(n):
            if i != rank and aug[i, col] == 1:
                aug[i, :] ^= aug[rank, :]
        rank += 1

    ##  检查是否有解
    for i in range(rank, n):
        if aug[i, -1] == 1:
            return None  # 无解的情况

    ##  构造解向量
    GF=galois.GF(2)
    solution = GF.Zeros(m)
    leading_cols = []

    ##  找出主元列
    for i in range(rank):
        for j in range(m):
            if aug[i, j] == 1:
                leading_cols.append(j)
                break

    ##  回代求解
    for i in range(rank):
        col = leading_cols[i]
        solution[col] = aug[i, -1]

        ##  消去当前行中主元列右侧的元素对解的影响
        for j in range(col + 1, m):
            if aug[i, j] == 1:
                solution[col] ^= solution[j]

    ##  ---返回解向量---
    return solution


##  USER：---返回两个线性空间的交集---
def cap(matrix1, matrix2):
    """""
    input.matrix1：GF(2)矩阵，行张成子空间1
    input.matrix2：GF(2)矩阵，行张成子空间2
    output：GF(2)矩阵，表示交集的行空间基
    """""

    ##  ---数据预处理---
    m = matrix1.shape[0]
    k = matrix2.shape[0]

    ##  解方程 basis1^T * x + basis2^T * y = 0 (表示交集向量), 构造矩阵 [basis1^T | basis2^T]
    aug_matrix = galois.GF(2)(np.concatenate((matrix1.T, matrix2.T), axis=1))

    ##  计算零空间（解空间）
    nullspace = aug_matrix.null_space()

    ##  从零空间中取对应于每个解向量的系数（对应于方程组中的解向量）
    ab_space = nullspace[:, :m]

    ##  将系数乘以 basis1 得到具体解（交集）
    if len(ab_space) == 0:
        return galois.GF(2).Zeros((0, matrix1.shape[1]))

    intersection_vectors = ab_space @ matrix1

    ##  产生基向量的线性无关组合，即每种向量的不同的高斯行组合样本
    rref_intersection = intersection_vectors.row_reduce()

    ##  除去全0行的行（即为没有有用基底捕获的时候）
    nz_mask = np.any(rref_intersection != 0, axis=1)
    rref_basis = rref_intersection[nz_mask]

    ##  返回补集
    return rref_basis


##  USER：---返回两个线性空间的差集（matrix1 \ cap(matrix1,matrix2)）---
def minus(matrix1, matrix2):
    """""
    input.matrix1：GF(2)矩阵，行张成子空间1
    input.matrix2：GF(2)矩阵，行张成子空间2
    output：列表[ndarray]，为子空间差的基向量（01行向量）
    """""

    ##  ---数据预处理---
    assert matrix1.shape[1] == matrix2.shape[1]

    ##  转化为GF2数组
    matrix1_origin = matrix1
    matrix2_origin = matrix2
    matrix1 = matrix1._array
    matrix2 = matrix2._array

    intersect = cap(matrix1_origin, matrix2_origin)._array
    if len(intersect) == 0:
        return galois.GF(2)(matrix1)

    result = []
    for i in range(len(matrix1)):
        rank = np.linalg.matrix_rank(intersect)
        intersect = np.vstack((intersect, matrix1[i]))
        if np.linalg.matrix_rank(intersect) > rank:
            result.append(matrix1[i])

    ##  返回差集
    return result


##  USER：---返回两个线性空间的直和基（合并并去冗余）---
def direct_sum(matrix1, matrix2):
    """""
    input.matrix1：GF(2)矩阵
    input.matrix2：GF(2)矩阵
    output：GF(2)矩阵，直和后的独立基
    """""
    assert isinstance(matrix1, galois.GF(2))
    assert isinstance(matrix2, galois.GF(2))

    ##  转化为GF2数组
    result = matrix1[0].copy()

    ##  直接拼接
    for i in range(1, len(matrix1)):
        rank = np.linalg.matrix_rank(result)
        temp = np.vstack((result, matrix1[i]))
        if np.linalg.matrix_rank(temp) > rank:
            result = temp
    for i in range(len(matrix2)):
        rank = np.linalg.matrix_rank(result)
        temp = np.vstack((result, matrix2[i]))
        if np.linalg.matrix_rank(temp) > rank:
            result = temp

    ##  返回结果
    return galois.GF(2)(result)


##  USER：---正交化基矢组（相对于标准内积）---
def orthogonalize(matrix):
    """""
    input.matrix：可迭代的GF(2)行向量列表/矩阵
    output：列表[GF(2)行向量]，两两正交的基
    """""

    ##  双线性形式矩阵
    B_i=[v.copy() for v in matrix]
    ortho_basis=[]  # 存储正交基

    while True:

        ##  遍历B_i，找到奇数权重向量
        length=len(B_i)
        for i in range(length):
            if np.mod(np.count_nonzero(B_i[i]), 2)==1:
                for j in range(length):
                    if j!=i and np.mod(np.count_nonzero(B_i[j]), 2)==0:
                        B_i[j]=B_i[j]+B_i[i]
                break
        flag=0
        b1=B_i[0]
        o_i=b1
        ortho_basis.append(o_i)
        next_B=[]
        for j in range(len(B_i)):
            if j!=flag:
                b=B_i[j]
                coef=np.dot(b, o_i)
                b_new=b+coef*o_i
                next_B.append(b_new)
        B_i=next_B
        if len(B_i)==0:
            break

    ##  ---返回正交基---
    return ortho_basis


##  USER：---计算code distance---
def distance(H, method:str,search_number:int=500)->int:
    """""
    input.H：GF(2)矩阵，稳定子或校验矩阵
    input.method：'mip' 或 'random'
    input.search_number：随机法搜索次数
    output：int，最小码距估计
    """""
    if method == 'mip':
        logic_op = minus(H.null_space(), H)
        return mip_distance_caculator(H, logic_op)
    elif method == 'random':
        return random_distance_caculator(H, H, search_number)
    else:
        raise NotImplementedError


#%%  KEY：===计算code distance===
def mip_distance_caculator(H, logicOp)->int:
    """""
    input.H：01矩阵或GF(2)矩阵
    input.logicOp：01矩阵或GF(2)矩阵，逻辑算子候选
    output：int，最小权重
    """""

    ##  格式化输入
    H = np.array(H, dtype=int)  # 转换为整数类型的numpy数组
    logicOp = np.array(logicOp, dtype=int)  # 转换为整数类型的numpy数组
    d = H.shape[1]  # 初始化距离为量子比特数量（最大可能距离）

    ##  遍历每个逻辑算子
    for i in range(logicOp.shape[0]):
        logicOp_i = logicOp[i, :]
        n = H.shape[1]  # 量子比特数量（稳定子矩阵的列数）
        m = H.shape[0]  # 稳定子数量（稳定子矩阵的行数）
        wstab = np.max([np.sum(H[i, :]) for i in range(m)])  # 计算最大稳定子权重（单个稳定子中非零元素的最大数量）
        wlog = np.count_nonzero(logicOp_i)  # 计算逻辑算子的权重
        num_anc_stab = int(np.ceil(np.log2(wstab)))  # 计算稳定子约束所需的辅助变量数量（基于最大稳定子权重的对数）
        num_anc_logical = int(np.ceil(np.log2(wlog)))  # 计算逻辑算子约束所需的辅助变量数量（基于逻辑算子权重的对数）
        num_var = n + m * num_anc_stab + num_anc_logical  # 总变量数量 = 量子比特变量 + 稳定子辅助变量 + 逻辑算子辅助变量

        ##  创建混合整数规划模型
        model = Model()
        model.verbose = 0  # 关闭详细输出
        x = [model.add_var(var_type=BINARY) for i in range(num_var)]  # 创建二进制变量数组
        model.objective = minimize(xsum(x[i] for i in range(n)))  # 目标函数：最小化前n个变量（量子比特变量）的和（即最小化Hamming权重）

        # 为每个稳定子添加正交性约束（模2）
        for row in range(m):
            weight = [0] * num_var  # 初始化权重向量
            supp = np.nonzero(H[row, :])[0]  # 获取当前稳定子的支持集（非零元素的位置）

            ##  设置qubit变量的权重为1
            for q in supp:
                weight[q] = 1

            ##  添加辅助变量来处理模2约束
            cnt = 1
            for q in range(num_anc_stab):
                weight[n + row * num_anc_stab + q] = -(1 << cnt)  # 设置辅助变量的权重为负的2的幂次方
                cnt += 1
            model += xsum(weight[i] * x[i] for i in range(num_var)) == 0  # 添加约束：权重向量与变量向量的点积等于0

        ##  添加逻辑算子的奇数重叠约束
        supp = np.nonzero(logicOp_i)[0]  # 获取逻辑算子的支持集
        weight = [0] * num_var  # 初始化权重向量

        ##  设置qubit变量的权重为1
        for q in supp:
            weight[q] = 1

        ##  添加辅助变量来处理模2约束
        cnt = 1
        for q in range(num_anc_logical):
            # 设置辅助变量的权重为负的2的幂次方
            weight[n + m * num_anc_stab + q] = -(1 << cnt)
            cnt += 1

        ##  添加约束：权重向量与变量向量的点积等于1（奇数重叠）
        model += xsum(weight[i] * x[i] for i in range(num_var)) == 1

        ##  求解优化问题，计算最优解中前n个变量的和，即最小Hamming权重
        model.optimize()
        # noinspection PyTypeChecker
        opt_val = sum([x[i].x for i in range(n)])
        d = min(d, int(opt_val))

    ##  返回最小距离
    return d


#%%  KEY：===计算随机距离===
def random_distance_caculator(gx, gz, num)->int:
    """""
    input.gx：GF(2)矩阵
    input.gz：GF(2)矩阵
    input.num：搜索迭代次数
    output：int，随机启发式最小权重
    """""

    ##  设置默认有限域为GF(2)（二进制域）
    F = galois.GF(2)

    wz = F(gx.null_space())  # 计算X稳定子的零空间（Z逻辑算子空间）
    wx = F(gz.null_space())  # 计算Z稳定子的零空间（X逻辑算子空间）
    rows_wz, cols_wz = wz.shape  # 获取零空间矩阵的维度信息
    dist_bound = cols_wz + 1  # 初始化距离上界为最大可能值（列数+1）
    vec_count = 0  # 计数器：记录找到当前最小权重的向量数量

    ##  主循环：进行num次随机迭代
    for i in range(num):
        per = np.random.permutation(cols_wz)  # 生成随机排列，用于随机化搜索顺序
        wz1 = wz[:, per]  # 对零空间矩阵的列进行随机排列
        wz2 = wz1.row_reduce()  # 对排列后的矩阵进行行约简（高斯消元）
        wz2 = wz2[:, np.argsort(per)]  # 将列顺序恢复为原始顺序

        ##  遍历行约简后的每一行
        for j in range(rows_wz):
            temp_vec = wz2[j, :]  # 获取当前行向量
            temp_weight = np.count_nonzero(temp_vec)  # 计算向量的Hamming权重（非零元素个数）

            ##  检查权重是否在有效范围内且小于等于当前最小距离
            if 0 < temp_weight <= dist_bound:

                ##  检查向量是否与X逻辑算子空间有非零重叠（即是否为非平凡逻辑算子）
                if np.any(wx @ temp_vec):

                    ##  如果找到更小的权重，更新最小距离
                    if temp_weight < dist_bound:
                        dist_bound = temp_weight
                        vec_count = 1

                    ##  如果权重等于当前最小距离，增加计数器
                    elif temp_weight == dist_bound:
                        vec_count += 1

            ##  检查是否达到最小距离阈值，如果达到则提前终止
            if dist_bound <= 2:
                return 2

    ##  返回找到的最小距离
    return dist_bound


##  USER：---生成循环移位矩阵---
def shift(number:'int', shift:'int'):
    """""
    input.number：矩阵维度
    input.shift：循环移位步长
    output：GF(2)矩阵，置换矩阵
    """""

    S = np.zeros((number, number), dtype=int)
    for i in range(number):
        S[i, (i + shift) % number] = 1
    return galois.GF(2)(S)


##  USER：---返回向量中为1的下标---
def occupy(vector):
    """""
    input.vector：01向量或GF(2)向量
    output：ndarray，值为1的位置索引
    influence：从向量表示提取支撑集
    """""

    return np.where(vector==1)[0]



"""""
模拟现象级噪声下Majorana Steane code的性能，code在d=20的时候近似达到了d_max=3.
"""""
from multiprocessing import Pool
import numpy as np
from extendedstim import MajoranaCSSCode, Circuit, Code2Circuit


def calculate(circuit,p):
    circuit.noise_amplitude_fix(p, p / 10)
    return circuit.experiment(int(np.floor(100/p)), 'tesseract')


def steane_code(d):
    code=MajoranaCSSCode.SteaneCode()
    circuit:Circuit=Code2Circuit(code,0,0,'phenomenological',d)
    physical_error_rates=np.linspace(1e-3, 1e-2, 20)
    _=circuit.detector_error_model()
    print(_)
    with Pool(processes=len(physical_error_rates)) as pool:
        results=[pool.apply_async(calculate, args=(circuit.copy(),p_temp)) for p_temp in physical_error_rates]
        final_results=[result.get() for result in results]
    logical_failure_rate=np.array(final_results)
    sample_numbers=np.floor(100/physical_error_rates)
    vars=2*3.2905*(logical_failure_rate-logical_failure_rate**2)/np.sqrt(sample_numbers)
    print('vars:', ','.join([str(temp) for temp in vars]))
    print('physical error rate:', ','.join([str(temp) for temp in physical_error_rates]))
    print('logical error rate:', ", ".join([str(temp) for temp in final_results]))
    file=open(f'steane_{d}.txt', 'w')
    file.write('physical error rate:'+','.join([str(temp) for temp in physical_error_rates])+'n')
    file.write('logical error rate:'+', '.join([str(temp) for temp in final_results])+'n')
    file.write('vars:'+','.join([str(temp) for temp in vars])+'n')
    file.close()


if __name__ == '__main__':
    steane_code(20)


"""""
模块作用：定义CSS量子码抽象基类，区分X/Z两类生成元并提供相应的秩、校验矩阵与距离抽象接口。
"""""
from abc import abstractmethod
import numpy as np
from extendedstim.Code.QuantumCode.QuantumCode import QuantumCode
from extendedstim.Physics.Operator import Operator


class QuantumCSSCode(QuantumCode):

    #%%  USER：构造方法
    def __init__(self, generators_x, generators_z, physical_number):
        """""
        input.generators_x：X类稳定子
        input.generators_z：Z类稳定子
        input.physical_number：物理位数
        """""
        self.generators_x = np.array(generators_x,dtype=Operator)
        self.generators_z = np.array(generators_z,dtype=Operator)
        self.checker_number_x=len(generators_x)
        self.checker_number_z = len(generators_z)
        self._logical_operators_x=None
        self._logical_operators_z=None
        super().__init__(generators_x+generators_z, physical_number)

    #%%  USER：属性方法
    ##  USER：求逻辑算符（X方向）
    @property
    @abstractmethod
    def logical_operators_x(self):
        """""
        output：X类逻辑算符数组
        """""
        pass

    ##  USER：求逻辑算符（Z方向）
    @property
    @abstractmethod
    def logical_operators_z(self):
        """""
        output：Z类逻辑算符数组
        """""
        pass

    ##  USER：求校验矩阵的秩（X方向）
    @property
    def rank_x(self):
        """""
        output：int，rank(Hx)
        """""
        return self.check_matrix_x.rank

    ##  USER：求校验矩阵的秩（Z方向）
    @property
    def rank_z(self):
        """""
        output：int，rank(Hz)
        """""
        return self.check_matrix_z.rank

    ##  USER：求校验矩阵（X方向）
    @property
    def check_matrix_x(self):
        """""
        output：GF(2)矩阵，生成元X的占据向量的偶下标子向量
        """""
        matrix=Operator.get_matrix(self.generators_x,self.physical_number)
        matrix=matrix[:,0::2]
        return matrix

    ##  USER：求校验矩阵（Z方向）
    @property
    def check_matrix_z(self):
        """""
        output：GF(2)矩阵，生成元Z的占据向量的奇下标子向量
        """""
        matrix=Operator.get_matrix(self.generators_z,self.physical_number)
        matrix = matrix[:, 1::2]
        return matrix

    ##  USER：求码距（X方向）
    @property
    @abstractmethod
    def distance_x(self):
        """""
        output：int，X向最小非平凡逻辑权重
        """""
        pass

    ##  USER：求码距（Z方向）
    @property
    @abstractmethod
    def distance_z(self):
        pass



"""""
模块作用：实现Double-Bicycle LDPC码构造与若干已知优良参数的便捷生成函数。
"""""
import galois
import numpy as np
from extendedstim.Code.LinearCode.LinearCode import LinearCode
from extendedstim.tools.GaloisTools import shift, solve, distance


class DoubleBicycleCode(LinearCode):
    def __init__(self,n:int,a_occupy:list,b_occupy:list)->None:
        """""
        input.n：循环矩阵的大小
        input.a_occupy：循环矩阵A中1的位置
        input.b_occupy：循环矩阵B中1的位置
        output：无（构造对象）
        """""
        self.a_occupy=a_occupy.copy()
        self.b_occupy=b_occupy.copy()
        self.a_occupy.sort()
        self.b_occupy.sort()
        a_list = [1 if i in self.a_occupy else 0 for i in range(n)]
        b_list = [1 if i in self.b_occupy else 0 for i in range(n)]
        H_up = np.hstack([C(n, a_list), C(n, b_list), C(n, a_list).T, C(n, b_list).T])
        H_down = np.hstack([C(n, b_list).T, C(n, a_list).T, C(n, b_list), C(n, a_list)])
        H = np.vstack([H_up, H_down])
        super().__init__(H)

    #%%  USER：===重载运算符===
    ##  USER：----打印代码----
    def __str__(self):
        n=self.number_bit//4
        S_str_a=''
        for i in self.a_occupy:
            S_str_a+=f"S^{{{i}}}_{{{n}}}"
        S_str_b=''
        for i in self.b_occupy:
            S_str_b+=f"S^{{{i}}}_{{{n}}}"
        return S_str_a+ '\n'+S_str_b

    #%%  USER：===静态方法===
    ##  USER：----生成好的代码----
    @staticmethod
    def good_code(n,*args):
        """""
        input.n：循环尺寸
        output：DoubleBicycleCode 或 None
        """""
        if n==23:  # d=14,k=4
            return DoubleBicycleCode(23,[0,9],[21,22])
        elif n==11:  # d=8,k=4
            return DoubleBicycleCode(11,[5,7],[2,7])
        elif n==17:  # d=10,k=4
            return DoubleBicycleCode(17,[2,4],[9,16])
        elif n==36:  # d=12,k=8
            return DoubleBicycleCode(36,[2,19],[1,12])
        elif n==15 and args[0]==6:
            return DoubleBicycleCode(15,[12],[8,10])
        else:
            return None

##  USER：===构造多项式的变量元===
def S(n):
    """""
    input.n：尺寸
    output：GF(2)置换矩阵 S (1步移位)
    """""
    return shift(n,1)


##  USER：===构造cyclic matrix===
def C(n,a_list):
    """""
    input.n：尺寸
    input.a_list：长度为n的01序列，指示多项式系数
    output：GF(2)循环矩阵 C(n,a)
    """""
    assert len(a_list) == n
    return galois.GF2(np.sum([a*(np.linalg.matrix_power(S(n),i)) for i,a in enumerate(a_list)],axis=0))
